# Copyright (c) 2015 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
from __future__ import absolute_import

try:
    import simplejson as json
except ImportError:
    import json

import sys
import codecs
from functools import partial
import logging

import pyschema

from pyschema_extensions import avro  # import to get avro mixins for fields
from pyschema import source_generation
assert avro  # silence linter

logger = logging.getLogger(__name__)

SIMPLE_FIELD_MAP = {
    "string": pyschema.Text,
    "float": partial(pyschema.Float, size=4),
    "double": pyschema.Float,
    "int": partial(pyschema.Integer, size=4),
    "boolean": pyschema.Boolean,
    "long": pyschema.Integer,
    "bytes": pyschema.Bytes
}


class AVSCParseException(Exception):
    pass


def parse_schema_string(schema_string):
    """
    Load and return a PySchema class from an avsc string
    """
    if isinstance(schema_string, str):
        schema_string = schema_string.decode("utf8")
    schema_struct = json.loads(schema_string)

    return AvroSchemaParser().parse_schema_struct(schema_struct)


class AvroSchemaParser(object):

    def __init__(self):
        self.schema_store = pyschema.core.SchemaStore()

    def _parse_schema_or_enum_struct(self, struct):
        # experimental interface...
        field = self._get_field_builder(struct, None)()
        if isinstance(field, pyschema.SubRecord):
            return field._schema
        elif isinstance(field, pyschema.Enum):
            self.schema_store.add_enum(field)
            return field
        else:
            raise AVSCParseException(
                "pyschema's avro parser doesn't support base schemas other than record and enum"
            )

    def parse_schema_struct(self, schema_struct, enclosing_namespace=None):
        record_name = schema_struct["name"]
        field_dct = {}
        namespace = None

        if "namespace" in schema_struct:
            namespace = schema_struct["namespace"]
        else:
            namespace = enclosing_namespace

        if namespace is not None:
            field_dct["_namespace"] = namespace

        if "fields" not in schema_struct:
            logger.warning("No fields in schema {}, skipping.".format(schema_struct))
            return None  # raise Exception("no fields in {}".format(schema_struct))

        for field_def in schema_struct["fields"]:
            field_name = field_def["name"]
            field_builder = self._get_field_builder(
                field_def["type"],
                namespace
            )

            if "default" in field_def:
                if field_def["default"] is not None:
                    default_parser = field_builder(nullable=False).avro_load
                    default_value = default_parser(field_def["default"])
                else:
                    default_value = None
            else:
                default_value = pyschema.core.NO_DEFAULT

            field = field_builder(
                description=field_def.get("doc"),
                default=default_value
            )
            field_dct[field_name] = field

        field_dct["__module__"] = "__avro_parser_runtime__"  # not great, but better than "abc"

        if "doc" in schema_struct:
            field_dct["__doc__"] = schema_struct["doc"]

        schema = pyschema.core.PySchema(record_name.encode("ascii"), (pyschema.core.Record,), field_dct)
        self.schema_store.add_record(schema)
        return schema

    def _get_field_builder(self, type_def_struct, enclosing_namespace):
        if isinstance(type_def_struct, list):
            return self._parse_union(type_def_struct, enclosing_namespace)
        elif isinstance(type_def_struct, dict):
            return self._parse_complex(type_def_struct, enclosing_namespace)
        elif type_def_struct in SIMPLE_FIELD_MAP:
            field_builder = SIMPLE_FIELD_MAP.get(type_def_struct)
            return partial(field_builder, nullable=False)
        elif isinstance(type_def_struct, basestring):
            return self._parse_reference(type_def_struct, enclosing_namespace)
        else:
            raise AVSCParseException((
                "Unknown type, and no prior declaration of: {0!r}"
            ).format(type_def_struct))

    def _parse_union(self, union_struct, enclosing_namespace):
        filtered = [subtype for subtype in union_struct if subtype != "null"]
        if len(filtered) > 1:
            raise AVSCParseException(
                "Unions with multiple non-null as type are unsupported: {0}".format(union_struct)
            )
        nullable = "null" in union_struct
        actual_type = filtered[0]
        field_builder = self._get_field_builder(actual_type, enclosing_namespace)

        return partial(field_builder, nullable=nullable)

    def _parse_map(self, type_def_struct, enclosing_namespace):
        values_field_builder = self._get_field_builder(
            type_def_struct["values"],
            enclosing_namespace
        )
        return partial(
            pyschema.Map,
            # NOTE: values field builder is executed here, so the field index
            # will be created at this point, but that shouldn't matter for the map type
            value_type=values_field_builder(),
            nullable=False
        )

    def _parse_reference(self, type_def_struct, enclosing_namespace):
        reference_name = type_def_struct
        if "." not in reference_name:
            # try first with added enclosing namespace
            full_name = ".".join([enclosing_namespace, reference_name])
            if self.schema_store.has_enum(full_name):
                enum_field_values = self.schema_store.get_enum(full_name)
                return partial(
                    pyschema.Enum,
                    name=full_name,
                    values=enum_field_values
                )
        if self.schema_store.has_enum(reference_name):
            enum_field_values = self.schema_store.get_enum(reference_name)
            return partial(
                pyschema.Enum,
                name=reference_name,
                values=enum_field_values
            )
        elif self.schema_store.has_schema(reference_name):
            schema_class = self.schema_store.get(reference_name)
            return partial(
                pyschema.SubRecord,
                schema_class,
                nullable=False
            )
        else:
            raise AVSCParseException((
                "A schema type '{0!r}' was referenced"
                " without prior declaration as either a record or an enum."
            ).format(reference_name))

    def _parse_subrecord(self, type_def_struct, enclosing_namespace):
        if isinstance(type_def_struct, dict):
            if "fields" not in type_def_struct:
                raise AVSCParseException((
                    "No 'fields' definition found in subrecord"
                    " declaration: {0!r}"
                ).format(type_def_struct))
            schema_class = self.parse_schema_struct(type_def_struct, enclosing_namespace)

        return partial(
            pyschema.SubRecord,
            schema_class,
            nullable=False
        )

    def _parse_array(self, type_def_struct, enclosing_namespace):
        item_type = self._get_field_builder(type_def_struct["items"], enclosing_namespace)()
        return partial(
            pyschema.List,
            item_type,
            nullable=False
        )

    def _parse_enum(self, type_def_struct, enclosing_namespace):
        # copy and make a tuple just to ensure it isn't modified later
        values = tuple(type_def_struct["symbols"])
        if "namespace" in type_def_struct:
            name = ".".join([type_def_struct["namespace"], type_def_struct["name"]])
        else:
            name = type_def_struct["name"]

        builder = partial(
            pyschema.Enum,
            values,
            name=name,
            nullable=False
        )

        def build_and_add_to_enum_store(*args, **kwargs):
            field = builder(*args, **kwargs)
            return self.schema_store.add_enum(field)
        return build_and_add_to_enum_store

    COMPLEX_MAPPING = {
        "map": _parse_map,
        "record": _parse_subrecord,
        "array": _parse_array,
        "enum": _parse_enum
    }

    def _parse_complex(self, type_def_struct, enclosing_namespace):
        typename = type_def_struct["type"]
        parser_func = self.COMPLEX_MAPPING.get(typename)
        if parser_func:
            return parser_func(self, type_def_struct, enclosing_namespace)
        elif typename in SIMPLE_FIELD_MAP:
            # hack to support the weird case where you double wrap a simple type
            return SIMPLE_FIELD_MAP[typename]
        raise AVSCParseException("Unknown complex type: {0}".format(type_def_struct))


def to_python_source(s):
    """Return a Python syntax declaration of the schemas contained in `s`"""
    schema = parse_schema_string(s)
    return source_generation.to_python_source([schema])

if __name__ == "__main__":
    utf8_stdin = codecs.getreader("utf8")(sys.stdin)
    schema_struct = json.load(utf8_stdin)
    schema = AvroSchemaParser().parse_schema_struct(schema_struct)
    print source_generation.to_python_source([schema])



import warnings

warnings.warn(
    "pyschema.contrib.avro_to_pyschema is deprecated and will be removed.\n"
    "Please use the pyschema_extensions.avro_to_pyschema package instead.",
    DeprecationWarning,
    stacklevel=2
)

import pyschema_extensions.avro_to_pyschema
from pyschema_extensions.avro_to_pyschema import *



import copy
from pyschema import types
import pyschema
import os
import sys
from collections import defaultdict

DEFAULT_INDENT = " " * 4


class SourceGenerationError(Exception):
    pass


def to_python_source(classes, indent=DEFAULT_INDENT):
    """Convert a set of pyschemas to executable python source code

    Currently supports all built-in types for basic usage.

    Notably not supported:
    * Maintaining class hierarchy
    * Methods, properties and non-field attributes
    * SELF-references
    """
    return header_source() + "\n" + classes_source(classes, indent)


RESERVED_KEYWORDS = [
    "and", "del", "from", "not", "while", "as", "elif",
    "global", "or", "with", "assert", "else", "if",
    "pass", "yield", "break", "except", "import",
    "print", "class", "exec", "in", "raise", "continue",
    "finally", "is", "return", "def", "for", "lambda", "try"
]


def make_safe(package_name):
    parts = package_name.split(".")
    for kw in RESERVED_KEYWORDS:
        while kw in parts:
            i = parts.index(kw)
            parts[i] = kw + "_"
    return ".".join(parts)


class PackageBuilder(object):
    def __init__(self, target_folder, parent_package, indent=DEFAULT_INDENT):
        self.target_folder = target_folder
        self.parent_package = parent_package
        self.indent = indent

    def get_namespace(self, schema):
        try:
            namespace = make_safe(schema._namespace)
        except AttributeError:
            namespace = None
        return namespace

    def get_namespace_clusters(self, all_classes):
        namespace_cluster = defaultdict(set)
        for c in all_classes:
            namespace = self.get_namespace(c)
            namespace_cluster[namespace].add(c)
        return namespace_cluster

    def format_definitions(self, classes):
        return "\n\n".join([_class_source(c, self.indent) for c in classes])

    def write_namespace_file(self, namespace, module_code):
        if not namespace:
            key = ['__init__']
        else:
            key = namespace.split('.')
        output_file = os.path.join(self.target_folder, *key) + '.py'
        output_dir = os.path.dirname(output_file)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        with open(output_file, 'w') as out_fn:
            out_fn.write(module_code)

    def write_init_files(self):
        def touch_init_file_in_folder(folder):
            path = os.path.join(folder, '__init__.py')
            if not os.path.exists(path):
                open(path, 'w').close()

        touch_init_file_in_folder(self.target_folder)
        for root, dirs, _ in os.walk(self.target_folder):
            for d in dirs:
                touch_init_file_in_folder(os.path.join(root, d))

    def format_imports(self, imported_classes):
        if not imported_classes:
            return "\n"
        imported_namespaces = self.get_namespace_clusters(imported_classes)
        lines = []
        for namespace, schemas in imported_namespaces.iteritems():
            if self.parent_package:
                if namespace:
                    module = "{}.{}".format(self.parent_package, namespace)
                else:
                    module = self.parent_package
            else:
                if not namespace:
                    module = "."
                else:
                    module = namespace
            class_part = ", ".join(s.__name__ for s in schemas)
            lines.append("from {} import {}".format(module, class_part))
        return "\n".join(lines) + "\n\n"

    def _get_namespace_prefixes(self, namespaces):
        prefixes = []
        for n in namespaces:
            if n:
                parts = n.split(".")
                prefixes.append(".".join(parts[:-1]))
        return prefixes

    def from_classes_with_refs(self, classes):
        class_graph = CachedGraphTraverser()

        all_classes = set(classes)
        for c in classes:
            referenced_schemas = class_graph.find_descendants(c)
            all_classes |= set(referenced_schemas)

        namespace_cluster = self.get_namespace_clusters(all_classes)
        parent_namespaces = self._get_namespace_prefixes(namespace_cluster.keys())
        ordered_schemas = class_graph.get_reference_ordered_schemas(all_classes)

        # Since we don't want to use the previous cached results we create a new instance
        # This CachedGraphTraverser will only keep one-child depth for its find_descendants
        child_only_class_graph = CachedGraphTraverser()
        for namespace, classes in namespace_cluster.iteritems():
            inlined_classes = [c for c in ordered_schemas if c in classes]
            imported_classes = set()

            for inlined in inlined_classes:
                direct_references = child_only_class_graph.find_descendants(inlined, max_depth=1)
                imported_classes |= set([c for c in direct_references if c not in inlined_classes])

            module_code = (
                header_source() +
                self.format_imports(imported_classes) +
                self.format_definitions(inlined_classes)
            )
            if namespace not in parent_namespaces:
                filename = namespace
            else:
                filename = "{}.{}".format(namespace, "__init__")

            self.write_namespace_file(filename, module_code)
            self.write_init_files()


def to_python_package(classes, target_folder, parent_package=None, indent=DEFAULT_INDENT):
    '''
    This function can be used to build a python package representation of pyschema classes.
    One module is created per namespace in a package matching the namespace hierarchy.

    Args:
        classes: A collection of classes to build the package from
        target_folder: Root folder of the package
        parent_package: Prepended on all import statements in order to support absolute imports.
            parent_package is not used when building the package file structure
        indent: Indent level. Defaults to 4 spaces
    '''
    PackageBuilder(target_folder, parent_package, indent).from_classes_with_refs(classes)


def classes_source(classes, indent=DEFAULT_INDENT):
    all_classes = set(classes)
    class_graph = CachedGraphTraverser()
    for c in classes:
        referenced_schemas = class_graph.find_descendants(c)
        all_classes |= set(referenced_schemas)

    ordered = class_graph.get_reference_ordered_schemas(all_classes)
    return "\n\n".join([_class_source(c, indent) for c in ordered])


def header_source():
    """Get the required header for generated source"""
    return (
        "import pyschema\n"
        "from pyschema.types import *\n"
        "from pyschema.core import NO_DEFAULT\n"
    )


def _class_source(schema, indent):
    """Generate Python source code for one specific class

    Doesn't include or take into account any dependencies between record types
    """

    def_pattern = (
        "class {class_name}(pyschema.Record):\n"
        "{indent}# WARNING: This class was generated by pyschema.to_python_source\n"
        "{indent}# there is a risk that any modification made to this class will be overwritten\n"
        "{optional_namespace_def}"
        "{field_defs}\n"
    )
    if hasattr(schema, '_namespace'):
        optional_namespace_def = "{indent}_namespace = {namespace!r}\n".format(
            namespace=schema._namespace, indent=indent)
    else:
        optional_namespace_def = ""

    field_defs = [
        "{indent}{field_name} = {field!r}".format(field_name=field_name, field=field, indent=indent)
        for field_name, field in schema._fields.iteritems()
    ]
    if not field_defs:
        field_defs = ["{indent}pass".format(indent=indent)]

    return def_pattern.format(
        class_name=schema._schema_name,
        optional_namespace_def=optional_namespace_def,
        field_defs="\n".join(field_defs),
        indent=indent
    )


class CachedGraphTraverser(object):
    def __init__(self):
        self.descendants = {}
        self.started = set()

    def find_descendants(self, a, max_depth=sys.getrecursionlimit()):
        if a in self.descendants:
            # fetch from cache
            return self.descendants[a]
        self.started.add(a)
        subs = set()
        if max_depth > 0:
            if pyschema.ispyschema(a):
                for _, field in a._fields.iteritems():
                    subs |= self.find_descendants(field, max_depth)
                self.descendants[a] = subs
            elif isinstance(a, types.List):
                subs |= self.find_descendants(a.field_type, max_depth)
            elif isinstance(a, types.Map):
                subs |= self.find_descendants(a.value_type, max_depth)
            elif isinstance(a, types.SubRecord):
                subs.add(a._schema)
                if a._schema not in self.started:  # otherwise there is a circular reference
                    subs |= self.find_descendants(a._schema, max_depth-1)
            self.started.remove(a)
        return subs

    def get_reference_ordered_schemas(self, schema_set):
        for schema in schema_set:
            self.find_descendants(schema)
        descendants = copy.deepcopy(self.descendants)  # a working copy

        ordered_output = []
        while descendants:
            leaves = []
            for root, referenced in descendants.iteritems():
                if len(referenced) == 0:
                    leaves.append(root)

            if not leaves:
                raise SourceGenerationError("Circular reference in input schemas, aborting")
            ordered_output += leaves
            for leaf in leaves:
                # remove all leaves
                descendants.pop(leaf)
                for root, referenced in descendants.iteritems():
                    if leaf in referenced:
                        referenced.remove(leaf)
        return ordered_output



# Copyright (c) 2013 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from unittest import TestCase
from pyschema import Record, dumps, loads, ispyschema, no_auto_store
from pyschema.types import *
import pyschema.core


class RevertDefinitionsTest(TestCase):
    def setUp(self):
        self._original_schemas = pyschema.core.auto_store  # original
        pyschema.core.auto_store = self._original_schemas.clone()

    def tearDown(self):
        pyschema.core.auto_store = self._original_schemas


class TestNestedRecord(RevertDefinitionsTest):
    def test_full_circle(self):
        class Foo(Record):
            bin = Bytes()

        class MyRecord(Record):
            a_string = Text()
            a_float = Float()
            record = List(SubRecord(Foo))

        rec = MyRecord(a_string=u"hej")
        rec.record = [Foo(bin="bar")]

        s = dumps(rec)
        reloaded_obj = loads(s)

        self.assertEquals(reloaded_obj.a_string, u"hej")
        self.assertTrue(reloaded_obj.a_float is None)
        self.assertTrue(reloaded_obj.record[0].bin, "bar")


class TestBaseRecordNotInStore(TestCase):
    def test(self):
        self.assertTrue(Record not in pyschema.core.auto_store)


class TestBasicUsage(TestCase):
    def setUp(self):
        @no_auto_store()
        class Foo(Record):
            t = Text()
            i = Integer()
            b = Boolean()

            def calculated(self):
                return self.t * 2

        self.Foo = Foo

    def test_class_field(self):
        record = self.Foo(t=u"foo")
        self.assertEquals(record.t, u"foo")

    def test_post_declaration_field(self):
        record = self.Foo(i=10)
        self.assertEquals(record.i, 10)

    def test_setattr_field(self):
        record = self.Foo()
        self.assertTrue(record.b is None)
        record.b = False
        self.assertFalse(record.b)

    def test_forbidden_assignment(self):
        record = self.Foo()

        def forbidden_assignment():
            record.c = "Something"

        self.assertRaises(
            AttributeError,
            forbidden_assignment
        )

    def test_record_name(self):
        self.assertEquals(
            self.Foo._schema_name,
            "Foo"
        )

    def test_type_adherence(self):
        self.assertTrue(ispyschema(self.Foo))
        self.assertTrue(issubclass(self.Foo, Record))
        self.assertTrue(isinstance(self.Foo, pyschema.core.PySchema))

    def test_method(self):
        record = self.Foo(t=u"a")
        calc = record.calculated()
        self.assertEquals(calc, u"aa")


class TestRuntimeRecord(TestBasicUsage):
    def setUp(self):
        class Foo(object):
            t = Text()

            def calculated(self):
                return self.t * 2

        Foo.i = Integer()
        setattr(Foo, "b", Boolean())
        self.Foo = pyschema.core.PySchema.from_class(Foo, auto_store=False)



# Copyright (c) 2013 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""Postgres style SQL generation based on pyschemas

Quite incomplete and still a work in progress
"""

import re

from pyschema.types import Integer, Text, Float, Boolean, Date, DateTime


Integer.pg_type = "INT"
Text.pg_type = "TEXT"
Float.pg_type = "FLOAT"
Boolean.pg_type = "BOOLEAN"
Date.pg_type = "DATE"
DateTime.pg_type = "TIMESTAMP WITHOUT TIME ZONE"


def camel_case_to_underscore(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def types(schema):
    all_types = []
    for name, field_type in schema._fields.iteritems():
        all_types.append((name, field_type.pg_type))
    return all_types


def _create_statement(table_name, types):
    parts = []
    for fielddef in types:
        parts.append("%s %s" % fielddef)
    coldefs = ", ".join(parts)
    return "CREATE TABLE %s (" % (table_name,) + coldefs + ")"


def create_statement(schema, table_name=None):
    table_name = table_name or camel_case_to_underscore(
        schema._schema_name)
    return _create_statement(
        table_name,
        types(schema)
    )



# Copyright (c) 2014 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the 'License'); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""Tools to convert from Avro schemas to PySchema schemas

Compatibility intended to be production-grade except:
* defaults;
* namespaces (not scrutinized);
* aliases;
* advanced Avro union types (other than the pattern to union with null, that
  translates to nullable i.e. optional).

This module doesn't deal at all with compatibility of data serialization logic
proper, only with schema definitions.
"""

import json

field_map = {
    'string': 'pyschema.Text',
    'float': 'pyschema.Float',
    'double': 'pyschema.Float',
    'int': 'pyschema.Integer',
    'boolean': 'pyschema.Boolean',
    'long': 'pyschema.Integer',
    'bytes': 'pyschema.Bytes',
}

extra_args_map = {
    'float': ['size=4'],
    'int': ['size=4'],
}

complex_field_map = {
    'array': 'pyschema.List',
    'map': 'pyschema.Map',
    'record': 'pyschema.SubRecord',
}


def get_ununionized_field_type(field_type):
    if isinstance(field_type, list):
        if len(field_type) != 2 or field_type[0] != 'null':
            raise NotImplementedError("PySchema doesn't support such advanced union types yet: {0!r}".format(field_type))
        return field_type[1]
    return field_type


def get_name(field):
    if isinstance(field['type'], basestring):
        return field['name']
    if isinstance(field['type'], list):
        return field['type'][0]['name']
    if isinstance(field['type'], dict):
        return field['type']['name']


def is_nullable(field_type):
    return isinstance(field_type, list) and 'null' in field_type


def get_field_type_name(field_type):
    field_type = get_ununionized_field_type(field_type)
    if isinstance(field_type, dict):
        field_type = field_type['type']
    return field_type


def get_sub_field_type(field):
    field_type = get_ununionized_field_type(field['type'])
    type_name = get_field_type_name(field['type'])
    if type_name == 'record':
        return field_type['fields']
    elif type_name == 'array':
        return field_type['items']
    elif type_name == 'map':
        return field_type['values']


def nullable_str(field_type):
    if not is_nullable(field_type):
        return 'nullable=False'
    return ''


def get_sub_fields_name(sub_type):
    sub_map = {'record': 'fields', 'array': 'items', 'map': 'values'}
    return sub_map[sub_type]


def get_sub_field(field):
    field_type = get_field_type_name(field['type'])
    if field_type == 'record':
        return field['fields']
    sub_field = field['type'][get_sub_fields_name(field_type)]
    if isinstance(sub_field, list):
        return sub_field[0]
    return sub_field


def get_field_definition(field, sub_records):
    args = []
    if isinstance(field, basestring):
        field_type_name = field
        args.append('nullable=False')
    else:
        field_type_name = get_field_type_name(field['type'])
        if is_nullable(field['type']):
            args.append('nullable=True')
        else:
            args.append('nullable=False')

    if 'doc' in field:
        args.append('description={0!r}'.format(field['doc'].encode('UTF-8')))

    # simple types
    if field_type_name in field_map:
        args.extend(extra_args_map.get(field_type_name, []))
        pyschema_type = field_map[field_type_name]
    # complex types
    elif field_type_name == 'record':
        args.insert(0, get_name(field))
        sub_rec = get_pyschema_record(field, sub_records)
        sub_records.append(sub_rec)
        pyschema_type = complex_field_map[field_type_name]
    else:
        sub_field = get_sub_field_type(field)
        args.insert(0, get_field_definition(sub_field, sub_records))
        pyschema_type = complex_field_map[field_type_name]
    return '{0}({1})'.format(pyschema_type, ', '.join(args))


def get_pyschema_record(schema, sub_records):
    if isinstance(schema, basestring):
        schema = json.loads(schema)

    name = get_name(schema)

    record_def = [
        "class {0}(pyschema.Record):".format(name),
        "# GENERATED BY {0}".format(__name__),
        "# YOU KNOW YOU WOULDN'T WANT IT ANY OTHER WAY",
        "# SO TAKE ME AS I AM",
    ]
    if 'doc' in schema:
        record_def += [repr(str(schema['doc']))]
    if 'namespace' in schema:
        record_def += ["_namespace = {0!r}".format(str(schema['namespace']))]

    if is_nullable(schema['type']):
        fields = schema['type'][0]['fields']
    else:
        fields = schema['fields']
    for field in fields:
        name = field['name']
        record_def += ["{0} = {1}".format(name, get_field_definition(field, sub_records))]
    return "\n    ".join(record_def) + "\n"



# Copyright (c) 2013 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
from __future__ import absolute_import

from pyschema.core import (
    PySchema, Record, dumps, loads, ispyschema, SchemaStore,
    disable_auto_register, enable_auto_register, no_auto_store,
    NO_DEFAULT
)
from pyschema.types import *
import warnings
warnings.simplefilter('default')



# Copyright (c) 2013 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
import datetime

import core
import copy
from core import ParseError, Field, auto_store, PySchema
import binascii
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


def ordereddict_push_front(dct, key, value):
    """Set a value at the front of an OrderedDict

    The original dict isn't modified, instead a copy is returned
    """
    d = OrderedDict()
    d[key] = value
    d.update(dct)
    return d


class Text(Field):
    def load(self, obj):
        if not isinstance(obj, (unicode, type(None))):
            raise ParseError("%r not a unicode object" % obj)
        return obj

    def dump(self, obj):
        if isinstance(obj, (unicode, type(None))):
            return obj
        else:
            try:
                return obj.decode('utf8')
            except:
                raise ValueError(
                    "%r is not a valid UTF-8 string" % obj
                )


class Bytes(Field):
    """Binary data"""

    def __init__(self, custom_encoding=False, **kwargs):
        super(Bytes, self).__init__(**kwargs)
        self.custom_encoding = custom_encoding

    def _load_utf8_codepoints(self, obj):
        return obj.encode("iso-8859-1")

    def _dump_utf8_codepoints(self, binary_data):
        return binary_data.decode("iso-8859-1")

    def _load_b64(self, obj):
        return binascii.a2b_base64(obj.encode("ascii"))

    def _dump_b64(self, binary_data):
        return binascii.b2a_base64(binary_data).rstrip('\n')

    def load(self, obj):
        if not self.custom_encoding:
            return self._load_utf8_codepoints(obj)
        return self._load_b64(obj)

    def dump(self, binary_data):
        if isinstance(binary_data, unicode):
            raise ValueError(
                "Unicode objects are not accepted values for Bytes (%r)"
                % (binary_data,)
            )
        if not self.custom_encoding:
            return self._dump_utf8_codepoints(binary_data)
        return self._dump_b64(binary_data)

    def is_similar_to(self, other):
        return super(Bytes, self).is_similar_to(other) and self.custom_encoding == other.custom_encoding


class List(Field):
    """List of one other Field type

    Differs from other fields in that it is not nullable
    and defaults to empty array instead of null
    """
    def __init__(self, field_type=Text(), nullable=False, default=[], **kwargs):
        super(List, self).__init__(nullable=nullable, default=default, **kwargs)
        self.field_type = field_type

    def load(self, obj):
        if not isinstance(obj, list):
            raise ParseError("%r is not a list object" % obj)
        return [self.field_type.load(o) for o in obj]

    def dump(self, obj):
        if not isinstance(obj, (tuple, list)):
            raise ValueError("%r is not a list object" % obj)
        return [self.field_type.dump(o) for o in obj]

    def set_parent(self, schema):
        self.field_type.set_parent(schema)

    def default_value(self):
        #  avoid default-sharing between records
        return copy.deepcopy(self.default)

    def is_similar_to(self, other):
        return super(List, self).is_similar_to(other) and self.field_type.is_similar_to(other.field_type)

    def repr_vars(self):
        return ordereddict_push_front(
            super(List, self).repr_vars(),
            "field_type",
            repr(self.field_type)
        )


class Enum(Field):
    _field_type = Text()  # don't change

    def __init__(self, values, name=None, **kwargs):
        super(Enum, self).__init__(**kwargs)
        self.values = set(values)
        self.name = name

        if name is not None and PySchema.auto_register:
            auto_store.add_enum(self)

    def dump(self, obj):
        if obj not in self.values:
            raise ValueError(
                "%r is not an allowed value of Enum%r"
                % (obj, tuple(self.values)))
        return self._field_type.dump(obj)

    def load(self, obj):
        parsed = self._field_type.load(obj)
        if parsed not in self.values and parsed is not None:
            raise ParseError(
                "Parsed value %r not in allowed value of Enum(%r)"
                % (parsed, tuple(self.values)))
        return parsed

    def is_similar_to(self, other):
        return super(Enum, self).is_similar_to(other) and self.values == other.values

    def repr_vars(self):
        return OrderedDict([
            ("values", self.values),
            ("name", repr(self.name))
        ] + super(Enum, self).repr_vars().items()
        )


class Integer(Field):
    def __init__(self, size=8, **kwargs):
        super(Integer, self).__init__(**kwargs)
        self.size = size

    def dump(self, obj):
        if not isinstance(obj, (int, long, type(None))) or isinstance(obj, bool):
            raise ValueError("%r is not a valid Integer" % (obj,))
        return obj

    def load(self, obj):
        if not isinstance(obj, (int, long, type(None))) or isinstance(obj, bool):
            raise ParseError("%r is not a valid Integer" % (obj,))
        return obj

    def is_similar_to(self, other):
        return super(Integer, self).is_similar_to(other) and self.size == other.size

    def repr_vars(self):
        return ordereddict_push_front(
            super(Integer, self).repr_vars(),
            "size",
            self.size
        )


class Boolean(Field):
    VALUE_MAP = {True: '1', 1: '1',
                 False: '0', 0: '0'}

    def dump(self, obj):
        if obj not in self.VALUE_MAP:
            raise ValueError(
                "Invalid value for Boolean field: %r" % obj)
        return bool(obj)

    def load(self, obj):
        if obj not in self.VALUE_MAP:
            raise ParseError(
                "Invalid value for Boolean field: %r" % obj)
        return bool(obj)


class Float(Field):
    def __init__(self, size=8, **kwargs):
        super(Float, self).__init__(**kwargs)
        self.size = size

    def dump(self, obj):
        if not isinstance(obj, float):
            raise ValueError("Invalid value for Float field: %r" % obj)
        return float(obj)

    def load(self, obj):
        if not isinstance(obj, (float, int, long)):
            raise ParseError("Invalid value for Float field: %r" % obj)
        return float(obj)

    def is_similar_to(self, other):
        return super(Float, self).is_similar_to(other) and self.size == other.size


class Date(Text):
    def dump(self, obj):
        if not isinstance(obj, datetime.date):
            raise ValueError("Invalid value for Date field: %r" % obj)
        return str(obj)

    def load(self, obj):
        try:
            return datetime.datetime.strptime(obj, "%Y-%m-%d").date()
        except ValueError:
            raise ValueError("Invalid value for Date field: %r" % obj)


class DateTime(Text):
    def dump(self, obj):
        if not isinstance(obj, datetime.datetime):
            raise ValueError("Invalid value for DateTime field: %r" % obj)
        return str(obj)

    def load(self, obj):
        try:
            if '.' in obj:
                return datetime.datetime.strptime(obj, "%Y-%m-%d %H:%M:%S.%f")
            return datetime.datetime.strptime(obj, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            raise ValueError("Invalid value for DateField field: %r" % obj)


# special value for SubRecord's schema parameter
# that signifies a SubRecord accepting records of the
# same type as the container/parent Record.
SELF = object()


class SubRecord(Field):
    """"Field for storing other :class:`record.Record`s"""

    def __init__(self, schema, **kwargs):
        super(SubRecord, self).__init__(**kwargs)
        self._schema = schema

    def dump(self, obj):
        if not isinstance(obj, self._schema):
            raise ValueError("%r is not a %r"
                             % (obj, self._schema))
        return core.to_json_compatible(obj)

    def load(self, obj):
        return core.from_json_compatible(self._schema, obj)

    def set_parent(self, schema):
        """This method gets called by the metaclass
        once the container class has been created
        to let the field store a reference to its
        parent if needed. Its needed for SubRecords
        in case it refers to the container record.
        """
        if self._schema == SELF:
            self._schema = schema

    def default_value(self):
        #  avoid default-sharing between records
        return copy.deepcopy(self.default)

    def is_similar_to(self, other):
        return super(SubRecord, self).is_similar_to(other) and self._schema == other._schema

    def repr_vars(self):
        return ordereddict_push_front(
            super(SubRecord, self).repr_vars(),
            "schema",
            self._schema._schema_name
        )


class Map(Field):
    """List of one other Field type

    Differs from other fields in that it is not nullable
    and defaults to empty array instead of null
    """
    def __init__(self, value_type, nullable=False, default={}, **kwargs):
        super(Map, self).__init__(nullable=nullable, default=default, **kwargs)
        self.value_type = value_type
        self.key_type = Text()

    def load(self, obj):
        return dict([
            (self.key_type.load(k),
             self.value_type.load(v))
            for k, v in obj.iteritems()
        ])

    def dump(self, obj):
        if not isinstance(obj, dict):
            raise ValueError("%r is not a dict" % (obj,))

        return dict([
            (self.key_type.dump(k),
             self.value_type.dump(v))
            for k, v in obj.iteritems()
        ])

    def set_parent(self, schema):
        self.value_type.set_parent(schema)

    def default_value(self):
        #  avoid default-sharing between records
        return copy.deepcopy(self.default)

    def is_similar_to(self, other):
        return super(Map, self).is_similar_to(other) and self.value_type.is_similar_to(other.value_type)

    def repr_vars(self):
        return ordereddict_push_front(
            super(Map, self).repr_vars(),
            "value_type",
            repr(self.value_type)
        )



# Copyright (c) 2014 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the 'License'); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
'''
Extension for generating JSON schema schemas from PySchema classes

JSON schema: http://json-schema.org/

When dumping to JSON schema, all fields in a record are mandatory, although
a list or map can be empty (but must be present). These records are still
dump-able, but they will not validate against the schema.

Usage:

>>> class MyRecord(pyschema.Record):
>>>     foo = Text()
>>>     bar = Integer()
>>>
>>> [pyschema_extensions.jsonschema.]get_root_schema_string(MyRecord)

'{"additionalProperties": false, "required": ["bar", "foo"], "type": "object", "id": "MyRecord", "properties": {"foo": {"t
ype": "string"}, "bar": {"type": "integer"}}}  '
'''

from pyschema import core
from pyschema.types import Field, Boolean, Integer, Float
from pyschema.types import Text, Enum, List, Map, SubRecord
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
try:
    import simplejson as json
except ImportError:
    import json


# Bytes are not supported
Boolean.jsonschema_type_name = 'boolean'
Integer.jsonschema_type_name = 'integer'
Float.jsonschema_type_name = 'number'
Text.jsonschema_type_name = 'string'
Enum.jsonschema_type_name = 'string'
List.jsonschema_type_name = 'array'
Map.avro_type_name = 'object'


@Field.mixin
class FieldMixin:
    def jsonschema_type_schema(self, state):
        return {
            'type': self.jsonschema_type_name,
        }


@Enum.mixin
class EnumMixin:
    def jsonschema_type_schema(self, state):
        return {
            'type': self.jsonschema_type_name,
            'enum': sorted(list(self.values)),
        }


@List.mixin
class ListMixin:
    def jsonschema_type_schema(self, state):
        return {
            'type': self.jsonschema_type_name,
            'items': self.field_type.jsonschema_type_schema(state),
        }


@Map.mixin
class MapMixin:
    def jsonschema_type_schema(self, state):
        return {
            'type': 'object',
            'additionalProperties': True,
            'patternProperties': {
                '^.*$': self.value_type.jsonschema_type_schema(state),
            },
        }


@SubRecord.mixin
class SubRecordMixin:
    def jsonschema_type_schema(self, state):
        cls = self._schema
        state.record_schemas[cls._schema_name] = get_schema_dict(cls, state)
        return {
            '$ref': self.jsonschema_type_name,
        }

    @property
    def jsonschema_type_name(self):
        return '#/definitions/{0}'.format(self._schema._schema_name)


# Schema generation
class SchemaGeneratorState(object):
    def __init__(self):
        self.record_schemas = dict()


def get_schema_dict(record, state=None):
    """Return a python dict representing the jsonschema of a record

    Any references to sub-schemas will be URI fragments that won't be
    resolvable without a root schema, available from get_root_schema_dict.
    """
    state = state or SchemaGeneratorState()
    schema = OrderedDict([
        ('type', 'object'),
        ('id', record._schema_name),
    ])
    fields = dict()
    for field_name, field_type in record._fields.iteritems():
        fields[field_name] = field_type.jsonschema_type_schema(state)

    required = set(fields.keys())
    schema['properties'] = fields
    schema['required'] = sorted(list(required))
    schema['additionalProperties'] = False

    state.record_schemas[record._schema_name] = schema
    return schema


def get_root_schema_dict(record):
    """Return a root jsonschema for a given record

    A root schema includes the $schema attribute and all sub-record
    schemas and definitions.
    """
    state = SchemaGeneratorState()
    schema = get_schema_dict(record, state)
    del state.record_schemas[record._schema_name]
    if state.record_schemas:
        schema['definitions'] = dict()
        for name, sub_schema in state.record_schemas.iteritems():
            schema['definitions'][name] = sub_schema
    return schema


def get_root_schema_string(record):
    return json.dumps(get_root_schema_dict(record))


def dumps(record):
    return json.dumps(core.to_json_compatible(record))


def loads(s, record_store=None, schema=None):
    return core.loads(s, record_store, schema, core.from_json_compatible)



# Copyright (c) 2013 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
Extension for generating Avro schemas from PySchema Record classes

Usage:

>>> class MyRecord(pyschema.Record):
>>>     foo = Text()
>>>     bar = Integer()
>>>
>>> [pyschema_extensions.avro.]get_schema_string(MyRecord)

'{"fields": [{"type": "string", "name": "foo"},
{"type": "long", "name": "bar"}],
"type": "record", "name": "MyRecord"}'

"""
import warnings
from pyschema import core
from pyschema.types import Field, Boolean, Integer, Float
from pyschema.types import Bytes, Text, Enum, List, Map, SubRecord
try:
    import simplejson as json
except ImportError:
    import json


Boolean.avro_type_name = "boolean"
Bytes.avro_type_name = "bytes"
Text.avro_type_name = "string"
# "ENUM" is the avro 'type name' of all enums generated by pyschema
# this is pyschema convention, not avro, so it might change if
# need b
Enum.avro_type_name = "ENUM"
List.avro_type_name = "array"
Map.avro_type_name = "map"


@Float.mixin
class FloatMixin:
    @property
    def avro_type_name(self):
        if self.size <= 4:
            return 'float'
        return 'double'


@Integer.mixin
class IntegerMixin:
    @property
    def avro_type_name(self):
        if self.size <= 4:
            return 'int'
        return 'long'


@Field.mixin
class FieldMixin:
    def avro_type_schema(self, state):
        """Full type specification for the field

        I.e. the same as would go into the "type" field.
        For most field, only simplified_avro_type_schema has
        to be implemented.
        """
        simple_type = self.simplified_avro_type_schema(state)
        if self.nullable:
            # first value in union needs to be same as default
            if self.default in (None, core.NO_DEFAULT):
                return ["null", simple_type]
            else:
                return [simple_type, "null"]
        else:
            return simple_type

    def simplified_avro_type_schema(self, state):
        """The basic avro type for this field

        Not including nullability.
        """
        return self.avro_type_name

    def avro_dump(self, o):
        if o is None:
            return None
        else:
            # relying on the reference json dump behavior
            # could be a bit dangerous
            if self.nullable:
                return {self.avro_type_name: self.dump(o)}
            else:
                return self.dump(o)

    def avro_load(self, o):
        if o is None:
            return None
        else:
            if self.nullable:
                return self.load(o[self.avro_type_name])
            else:
                return self.load(o)

    def avro_default_value(self):
        return self.default


@List.mixin
class ListMixin:
    def simplified_avro_type_schema(self, state):
        return {
            "type": "array",
            "items": self.field_type.avro_type_schema(state)
        }

    def avro_dump(self, obj):
        if obj is None:
            return None
        else:
            l = [self.field_type.avro_dump(o) for o in obj]
            if self.nullable:
                return {self.avro_type_name: l}
            else:
                return l

    def avro_load(self, obj):
        if obj is None:
            return None
        else:
            if self.nullable:
                obj = obj[self.avro_type_name]
            return [
                self.field_type.avro_load(o)
                for o in obj
            ]


### `Enum` extensions
@Enum.mixin
class EnumMixin:
    def simplified_avro_type_schema(self, state):
        return {
            "type": "enum",
            "name": self.avro_type_name,
            "symbols": list(self.values)
        }


@SubRecord.mixin
class SubRecordMixin:
    def simplified_avro_type_schema(self, state):
        return get_schema_dict(self._schema, state)

    @property
    def avro_type_name(self):
        return core.get_full_name(self._schema)

    def avro_dump(self, obj):
        if obj is None:
            return None
        if self.nullable:
            return {self.avro_type_name: to_json_compatible(obj)}
        else:
            return to_json_compatible(obj)

    def _get_record_data(self, obj):
        """
        Try to get the data for this subrecord using the full_name including namespace.
        If no data is given for the fullname, we will fallback to use the SubRecord name without namespace.

        Doing this allows namespace addition to a schema while being able to read legacy data missing a namespace.
        """
        if self.avro_type_name in obj:
            return obj[self.avro_type_name]
        else:
            base_name = self.avro_type_name.split('.')[-1]
            return obj[base_name]

    def avro_load(self, obj):
        if obj is None:
            return None
        if self.nullable:
            return from_json_compatible(
                self._schema,
                self._get_record_data(obj)
            )
        else:
            return from_json_compatible(
                self._schema,
                obj
            )


@Map.mixin
class MapMixin:
    def simplified_avro_type_schema(self, state):
        assert isinstance(self.key_type, Text)
        return {
            "type": "map",
            "values": self.value_type.avro_type_schema(state)
        }

    def avro_dump(self, obj):
        if obj is None:
            return None
        else:
            m = dict([(
                # using json loader for key is kind of a hack
                # since this isn't an actual type in avro (always text)
                self.key_type.dump(k),
                self.value_type.avro_dump(v)
            ) for k, v in obj.iteritems()])
            if self.nullable:
                return {self.avro_type_name: m}
            else:
                return m

    def avro_load(self, obj):
        if obj is None:
            return None
        else:
            if self.nullable:
                obj = obj[self.avro_type_name]
            m = dict([(
                # using json loader for key is kind of a hack
                # since this isn't an actual type in avro (always text)
                self.key_type.load(k),
                self.value_type.avro_load(v)
            ) for k, v in obj.iteritems()])
            return m


# Schema generation
class SchemaGeneratorState(object):
    def __init__(self):
        self.declared_records = set()


def get_schema_dict(record, state=None):
    state = state or SchemaGeneratorState()

    full_name = core.get_full_name(record)
    if '.' in full_name:
        namespace, record_name = full_name.rsplit('.', 1)
    else:
        namespace = None
        record_name = record._schema_name

    if record_name in state.declared_records:
        return record_name
    state.declared_records.add(record_name)

    avro_record = {
        "type": "record",
        "name": record._schema_name,
    }
    if namespace:
        avro_record["namespace"] = namespace

    if record.__doc__ is not None:
        avro_record["doc"] = record.__doc__

    avro_fields = []
    for field_name, field_type in record._fields.iteritems():
        field_spec = {
            "name": field_name,
            "type": field_type.avro_type_schema(state)
        }
        if field_type.default is not core.NO_DEFAULT:
            field_spec["default"] = field_type.avro_default_value()
        if field_type.description is not None:
            field_spec["doc"] = field_type.description
        avro_fields.append(field_spec)

    avro_record["fields"] = avro_fields
    return avro_record


def get_schema_string(record):
    return json.dumps(get_schema_dict(record))


def dumps(record):
    return json.dumps(to_json_compatible(record))


def to_json_compatible(record):
    dct = {}
    for name, fieldtype in record._fields.iteritems():
        value = getattr(record, name)
        dct[name] = fieldtype.avro_dump(value)
    return dct


def from_json_compatible(schema, dct):
    "Load from json-encodable"
    kwargs = {}

    for key in dct:
        field_type = schema._fields.get(key)
        if field_type is None:
            raise core.ParseError("Unexpected field encountered in line for record %s: %r" % (schema.__name__, key))
        kwargs[key] = field_type.avro_load(dct[key])

    return schema(**kwargs)


def loads(
    s,
    record_store=None,
    schema=None,
    record_class=None  # deprecated - replaced by `schema`
):
    if record_class is not None:
        warnings.warn(
            "The record_class parameter is deprecated in favour of schema",
            DeprecationWarning,
            stacklevel=2
        )
        schema = record_class

    return core.loads(s, record_store, schema, from_json_compatible)



# Copyright (c) 2013 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""Basic utilities for using PySchema in Luigi's Python MR lib
"""
import sys
from pyschema import core


def mr_reader(job, input_stream, loads=core.loads):
    """ Converts a file object with json serialised pyschema records
        to a stream of pyschema objects

    Can be used as job.reader in luigi.hadoop.JobTask
    """
    for line in input_stream:
        yield loads(line),


def mr_writer(job, outputs, output_stream,
              stderr=sys.stderr, dumps=core.dumps):
    """ Writes a stream of json serialised pyschema Records to a file object

    Can be used as job.writer in luigi.hadoop.JobTask
    """
    for output in outputs:
        try:
            print >> output_stream, dumps(output)
        except core.ParseError, e:
            print >> stderr, e
            raise



# Copyright (c) 2013 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from setuptools import setup

setup(
    name="pyschema",
    version="2.3.0",
    description="Schema definition and serialisation library",
    author="Elias Freider",
    author_email="freider@spotify.com",
    url="http://github.com/spotify/pyschema",
    packages=[
        "pyschema",
        "pyschema_extensions",
        "pyschema.contrib",  # deprecated package, replaced by pyschema_extensions
    ],
    keywords=["schema", "avro", "postgres", "json"],
    install_requires=[
        "simplejson"
    ],
    namespace_packages=[
        "pyschema_extensions"
    ],
    test_suite="nose.collector"
)



import warnings

warnings.warn(
    "pyschema.contrib.jsonschema is deprecated and will be removed.\n"
    "Please use the pyschema_extensions.jsonschema package instead.",
    DeprecationWarning,
    stacklevel=2
)

import pyschema_extensions.jsonschema
from pyschema_extensions.jsonschema import *



import warnings

warnings.warn(
    "pyschema.contrib.luigi is deprecated and will be removed.\n"
    "Please use the pyschema_extensions.luigi package instead.",
    DeprecationWarning,
    stacklevel=2
)

import pyschema_extensions.luigi
from pyschema_extensions.luigi import *



# Copyright (c) 2013 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.


""" Schema definition toolkit using Python classes

Usage example:

>>> class Foo(Record):
... bin = Bytes()
...
... class MyRecord(Record):
...     a_string = Text()
...     a_float = Float()
...     record = List(SubRecord(Foo))
...
... rec = MyRecord(a_string="hej")
... rec.record = [Foo(bin="bar")]
...
... s = dumps(rec)
... print loads(s)


Internals:

A valid PySchema class contains the following class variables:

`_fields`
    An OrderedDict of `field_name` => `field_type`
    where `field_type` is an instance of a Field subclass

`_schema_name`
    The qualifying name for this schema. This is used for registering a record
    in a `SchemaStore` and for auto-identification of serialized records.
    Should be unique within a specific SchemaStore, so if auto registering is
    used it should be unique within the execution chain of the current program.

"""
from __future__ import absolute_import
from abc import ABCMeta, abstractmethod
from itertools import izip

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import warnings
import types   # absolute import, this is the python standard library types
try:
    import simplejson as json
except ImportError:
    import json


SCHEMA_FIELD_NAME = "$schema"


class ParseError(Exception):
    """ Generic exception type for Record parse errors """
    pass


class SchemaStore(object):
    def __init__(self):
        self._schema_map = {}
        self._enum_map = {}

    def __str__(self):
        return str(self._schema_map.keys())

    def add_record(self, schema, _bump_stack_level=False):
        """ Add record class to record store for retrieval at record load time.

            Can be used as a class decorator
        """
        full_name = get_full_name(schema)
        self._force_add(full_name, schema, _bump_stack_level)
        if '.' in full_name and schema.__name__ not in self._schema_map:
            self._force_add(schema.__name__, schema, _bump_stack_level)
        return schema

    def add_enum(self, enum_definition):
        new_values_set = set(enum_definition.values)
        old_values_set = self._enum_map.get(enum_definition.name)

        if old_values_set is not None and new_values_set != old_values_set:
            warnings.warn(
                "Enum {!r} overwritten! Was: {}, Overwritten by: {}".format(
                    enum_definition.name,
                    old_values_set,
                    new_values_set
                )
            )

        if enum_definition.name is not None:
            self._enum_map[enum_definition.name] = enum_definition.values
        # return the definition to allow the method to be used as a decorator
        return enum_definition

    def _force_add(self, used_name, schema, _bump_stack_level=False):
        existing = self._schema_map.get(used_name, None)
        if existing:
            full_name = get_full_name(schema)
            explanation = "(actually {0})".format() if full_name != used_name else ""

            warnings.warn(
                "{used_name}{explanation}: old definition in {prev_module} replaced by definition in {new_module}"
                .format(used_name=used_name,
                        explanation=explanation,
                        prev_module=existing.__module__,
                        new_module=schema.__module__),
                stacklevel=4 if _bump_stack_level else 3)
        self._schema_map[used_name] = schema

    def get(self, record_name):
        """
        Will return a matching record or raise KeyError is no record is found.

        If the record name is a full name we will first check for a record matching the full name.
        If no such record is found any record matching the last part of the full name (without the namespace) will
        be returned.
        """
        if record_name in self._schema_map:
            return self._schema_map[record_name]
        else:
            last_name = record_name.split('.')[-1]
            return self._schema_map[last_name]

    def get_enum(self, name):
        return self._enum_map[name]

    def clear(self):
        self._schema_map.clear()
        self._enum_map.clear()

    def clone(self):
        r = SchemaStore()
        r._schema_map = self._schema_map.copy()
        r._enum_map = self._enum_map.copy()

        return r

    def has_schema(self, name):
        if name in self._schema_map:
            return True
        if "." in name:
            basename = name.split(".")[-1]
            return basename in self._schema_map
        return False

    def has_enum(self, name):
        return name in self._enum_map

    def __contains__(self, schema):
        return schema in self._schema_map.values()


class RecordStore(SchemaStore):
    def __init__(self):
        warnings.warn("RecordStore is deprecated and has been renamed to SchemaStore", DeprecationWarning, stacklevel=2)
        super(RecordStore, self).__init__()


def get_full_name(schema):
    full_name = schema.__name__
    if hasattr(schema, '_namespace'):
        full_name = '.'.join([schema._namespace, schema.__name__])
    elif hasattr(schema, '_avro_namespace_'):
        warnings.warn("_avro_namespace is deprecated, use _namespace instead", DeprecationWarning, stacklevel=3)
        full_name = '.'.join([schema._avro_namespace_, schema.__name__])
    return full_name


class _NoDefault:
    def __repr__(self):
        return "NO_DEFAULT"

# NO_DEFAULT is a special value to signify that a field has no default value
# and should fail to serialize unless a value has been assigned
# it's the default default-value for all non-nullable fields

NO_DEFAULT = _NoDefault()

_UNTOUCHED = object()


class Field(object):
    __metaclass__ = ABCMeta
    _next_index = 0

    def __init__(self, description=None, nullable=True, default=_UNTOUCHED):
        self.description = description
        self._index = Field._next_index
        self.nullable = nullable
        if default is _UNTOUCHED:
            # if default isn't explicitly set
            # use None for Nullables, and NO_DEFAULT for others
            if nullable:
                default = None
            else:
                default = NO_DEFAULT

        self.default = default
        Field._next_index += 1  # used for arg order in initialization

    def repr_vars(self):
        """Return a dictionary the field definition

        Should contain all fields that are required for the definition of this field in a pyschema class"""
        d = OrderedDict()
        d["nullable"] = repr(self.nullable)
        d["default"] = repr(self.default)
        if self.description is not None:
            d["description"] = repr(self.description)
        return d

    def __repr__(self):
        strings = ('{0}={1}'.format(vname, val) for vname, val in self.repr_vars().iteritems())
        return self.__class__.__name__ + '(' + ', '.join(strings) + ')'

    def set_parent(self, schema):
        # no-op by default but can be overridden by types
        # that need parent references
        pass

    @abstractmethod
    def dump(self, obj):
        pass

    @abstractmethod
    def load(self, obj):
        pass

    @classmethod
    def mixin(cls, mixin_cls):
        """Decorator for mixing in additional functionality into field type

        Example:

        >>> @Integer.mixin
        ... class IntegerPostgresExtensions:
        ...     postgres_type = 'INT'
        ...
        ...     def postgres_dump(self, obj):
        ...         self.dump(obj) + "::integer"

        Is roughly equivalent to:

        >>> Integer.postgres_type = 'INT'
        ...
        ... def postgres_dump(self, obj):
        ...     self.dump(obj) + "::integer"
        ...
        ... Integer.postgres_dump = postgres_dump

        """
        for item_name in dir(mixin_cls):
            if item_name.startswith("__"):
                # don't copy magic properties
                continue
            item = getattr(mixin_cls, item_name)

            if isinstance(item, types.MethodType):
                # unbound method will cause problems
                # so get the underlying function instead
                item = item.im_func

            setattr(cls, item_name, item)
        return mixin_cls

    def default_value(self):
        return self.default

    def is_similar_to(self, other):
        return(
            type(self) == type(other) and
            self.default == other.default and
            self.nullable == other.nullable and
            self.description == other.description
        )

auto_store = SchemaStore()


class PySchema(ABCMeta):
    """Metaclass for Records

    Builds schema on Record declaration and remembers Record types
    for easy generic parsing
    """
    auto_register = True

    def __new__(metacls, name, bases, dct):
        schema_attrs = metacls._get_schema_attributes(
            name=name,
            bases=bases,
            dct=dct
        )
        dct.update(schema_attrs)
        cls = ABCMeta.__new__(metacls, name, bases, dct)

        # allow self-references etc.
        for field_name, field in cls._fields.iteritems():
            field.set_parent(cls)

        if metacls.auto_register:
            auto_store.add_record(cls, _bump_stack_level=True)
        return cls

    @classmethod
    def _field_dupe_warning(metacls, name, fields):
        warnings.warn(
            "{schema}: Duplicate field definition for field{plural} {field}"
                .format(
                    schema=name,
                    field=fields,
                    plural="s" if len(fields) > 1 else ""
                ),
            stacklevel=4
        )

    @classmethod
    def _get_schema_attributes(metacls, name, bases, dct):
        fields = OrderedDict()
        for b in bases:
            if not isinstance(b, metacls):
                continue

            field_intersection = set(fields) & set(b._fields)
            if field_intersection:
                metacls._field_dupe_warning(name, field_intersection)
            fields.update(b._fields)

        new_fields = []
        for field_name, field_def in dct.iteritems():
            if isinstance(field_def, Field):
                new_fields.append((field_name, field_def))

        new_fields.sort(key=lambda fd: fd[1]._index)
        for field_name, field_def in new_fields:
            if field_name in fields:
                metacls._field_dupe_warning(name, (field_name,))
            fields[field_name] = field_def

        return {
            "_fields": fields,
            "_schema_name": name,
        }

    @classmethod
    def from_class(metacls, cls, auto_store=True):
        """Create proper PySchema class from cls

        Any methods and attributes will be transferred to the
        new object
        """
        if auto_store:
            def wrap(cls):
                return cls
        else:
            wrap = no_auto_store()

        return wrap(metacls.__new__(
            metacls,
            cls.__name__,
            (Record,),
            dict(cls.__dict__)
        ))


def disable_auto_register():
    PySchema.auto_register = False


def enable_auto_register():
    PySchema.auto_register = True


def no_auto_store():
    """ Temporarily disable automatic registration of records in the auto_store

    Decorator factory. This is _NOT_ thread safe

    >>> @no_auto_store()
    ... class BarRecord(Record):
    ...     pass
    >>> BarRecord in auto_store
    False

    """
    original_auto_register_value = PySchema.auto_register
    disable_auto_register()

    def decorator(cls):
        PySchema.auto_register = original_auto_register_value
        return cls

    return decorator


@no_auto_store()
class Record(object):
    """Abstract base class for structured logging records"""
    __metaclass__ = PySchema

    def __init__(self, *args, **kwargs):
        if args:
            # The idea behind only allowing keyword arguments
            # is to prevent accidental misuse of a changed schema
            raise TypeError('Non-keyword arguments not allowed'
                            ' when initializing Records')
        for k, field_type in self._fields.iteritems():
            object.__setattr__(self, k, field_type.default_value())
        for k, v in kwargs.iteritems():
            setattr(self, k, v)

    def __setattr__(self, name, value):
        if name not in self._fields:
            raise AttributeError(
                "No field %r in %s"
                % (name, self._schema_name)
            )

        super(Record, self).__setattr__(name, value)

    def __unicode__(self):
        return str(self).decode('ascii')

    def __str__(self):
        return repr(self)

    def __repr__(self):
        strings = ('%s=%r' % (fname, getattr(self, fname))
                   for fname, f in self._fields.iteritems())

        return self._schema_name + '(' + ', '.join(strings) + ')'

    def __cmp__(self, other):
        if not isinstance(other, Record):
            # return default implementation cmp value
            return cmp(id(self), other)
        if self._schema_name != other._schema_name:
            return cmp(self._schema_name, other._schema_name)
        fields = self._fields.keys()
        a = (getattr(self, key) for key in fields)
        b = (getattr(other, key) for key in fields)

        for _a, _b in izip(a, b):
            r = cmp(_a, _b)
            if r:
                return r
        return 0

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0


def to_json_compatible(record):
    "Dump record in json-encodable object format"
    d = {}
    for fname, f in record._fields.iteritems():
        val = getattr(record, fname)
        if val is not None:
            d[fname] = f.dump(val)
    return d


def from_json_compatible(schema, dct):
    "Load from json-encodable"
    kwargs = {}

    for key in dct:
        field_type = schema._fields.get(key)
        if field_type is None:
            raise ParseError("Unexpected field encountered in line for record %s: %s" % (schema.__name__, key))
        kwargs[key] = field_type.load(dct[key])

    return schema(**kwargs)


def ispyschema(schema):
    """ Is object PySchema instance?

    Returns true for PySchema Record *classes*
    i.e. NOT when schema is a *Record* instance

    >>> class FooRecord(Record):
    ...     pass
    >>> ispyschema(FooRecord)
    True
    >>> ispyschema(FooRecord())
    False
    """
    return isinstance(schema, PySchema)


def load_json_dct(
        dct,
        record_store=None,
        schema=None,
        loader=from_json_compatible
):
    """ Create a Record instance from a json-compatible dictionary

    The dictionary values should have types that are json compatible,
    as if just loaded from a json serialized record string.

    :param dct:
        Python dictionary with key/value pairs for the record

    :param record_store:
        Record store to use for schema lookups (when $schema field is present)

    :param schema:
        PySchema Record class for the record to load.
        This will override any $schema fields specified in `dct`

    """
    if schema is None:
        if record_store is None:
            record_store = auto_store
        try:
            schema_name = dct.pop(SCHEMA_FIELD_NAME)
        except KeyError:
            raise ParseError((
                "Serialized record missing '{0}' "
                "record identifier and no schema supplied")
                .format(SCHEMA_FIELD_NAME)
            )
        try:
            schema = record_store.get(schema_name)
        except KeyError:
            raise ParseError(
                "Can't recognize record type %r"
                % (schema_name,), schema_name)

    # if schema is explicit, use that instead of SCHEMA_FIELD_NAME
    elif SCHEMA_FIELD_NAME in dct:
        dct.pop(SCHEMA_FIELD_NAME)

    record = loader(schema, dct)
    return record


def loads(
        s,
        record_store=None,
        schema=None,
        loader=from_json_compatible,
        record_class=None  # deprecated in favor of schema
):
    """ Create a Record instance from a json serialized dictionary

    :param s:
        String with a json-serialized dictionary

    :param record_store:
        Record store to use for schema lookups (when $schema field is present)

    :param loader:
        Function called to fetch attributes from json. Typically shouldn't be used by end users

    :param schema:
        PySchema Record class for the record to load.
        This will override any $schema fields specified in `s`

    :param record_class:
        DEPRECATED option, old name for the `schema` parameter

    """
    if record_class is not None:
        warnings.warn(
            "The record_class parameter is deprecated in favour of schema",
            DeprecationWarning,
            stacklevel=2
        )
        schema = record_class
    if not isinstance(s, unicode):
        s = s.decode('utf8')
    if s.startswith(u"{"):
        json_dct = json.loads(s)
        return load_json_dct(json_dct, record_store, schema, loader)
    else:
        raise ParseError("Not a json record")


def dumps(obj, attach_schema_name=True):
    json_dct = to_json_compatible(obj)
    if attach_schema_name:
        json_dct[SCHEMA_FIELD_NAME] = get_full_name(obj.__class__)

    json_string = json.dumps(json_dct)
    return json_string
